
package Wiki::Page;
use Wiki::HTML;
use Wiki::Action;
use Text::Template;
use strict;
use Memoize;
use Carp;
use Errno 'EEXIST';
memoize 'filename';
memoize  'plugin';
our($ROOT, $PAGE, $PLUGIN);

sub import {
  (my $pack, $ROOT) = @_;
  unless (defined $ROOT) {
    require Carp;
    Carp::croak("Usage: use WikiPage 'rootpath'");
  }
  $ROOT =~ s|/+$||;
  $PAGE = "$ROOT/pages";
  $PLUGIN = "$ROOT/plugins";
}

sub filename {
  my $page = shift;
  $page =~ s|^/+||;
  $page =~ s|/+$||;
  $page =~ tr|.\0||d;
  "$PAGE/$page.wk";
}

sub exists {
  $Wiki::Page::plugins{$_[0]} || -f filename(@_);
}

BEGIN { 
  %Wiki::Page::plugins = ('RecentChanges' => "ChangeLog.plx",
                          'RecentEdits'   => "ChangeLog.plx",
                          'PageList'      => "PageList.plx",
                         );
}

sub plugin {
  my $page = shift;
  my $z = $Wiki::Page::plugins{$page};
  return unless $z;
  "$PLUGIN/$z";
}

sub contents {
  my $page = shift;
  my $file = filename($page);
  local *F;
  open F, "<", $file or return;
  local $/ ;
  return scalar <F>;
}

sub html {
  my $req = shift;
  my $page = $req->page;
  my $text = contents($page);
  return unless defined $text;
  $text = link_wikiwords($req, $text);
  scalar Wiki::HTML->new->do_text($text);
}

sub link_wikiwords {
  my $req = shift;
  my $text = shift;
  my @text = split /\n/, $text;
  for (@text) {
    next if /^    /;
    s{\b (
           (?: [A-Z] 
               (?>  [a-z0-9]+)
               /? )  {2,}
         )
         \b}{link_wikiword($req, $1)}gex;
    s{\b((?:(?>\w+)::)+\w+)\b}{link_wikiword($req, $1)}ge;
    s{\b(\w+\.pm)\b}{link_wikiword($req, $1)}ge;
  }
  join "\n", @text, "";
}

sub link_wikiword {
  my $req = shift;
  my $wikiword = shift;
  if (Wiki::Page::exists($wikiword)) {
    qq{W<$wikiword>};
  } else {
    "$wikiword" . "A<?|". requrl($req, "Create", $wikiword). ">";
  }
}

sub cur_rev {
  my $page = shift;
  my $file = filename($page);
  my $output = qx{rlog -h $file};
  my ($rev) = ($output =~ /^head: \d+\.(\d+)/m);
  $rev+0;
}

sub create {
  my ($req, $page) = @_;
  my $file = filename($page);
  mkdir_for($page);
  open F, ">>", $file or die "Couldn't create empty file $file for page $page: $!";
  close F;
  my $cmd = qq{ci -t-'Initial revision' -i $file};
  system($cmd) == 0 or die "rcs -i for $file yielded exit status $?" ;
  system("co -u $file") == 0 or die "co -u for $file yielded exit status $?" ;
  Log($req, $page, 1);
}

sub modify {
  my ($req, $page, $oldrev, $newtext) = @_;
  my $file = filename($page);
  my $result = `co -l $file 2>&1 </dev/null`;
  my $cur_rev = cur_rev($page);
  unless ($cur_rev eq $oldrev || !$oldrev) {
    warn "$page version skew: $oldrev submitted, $cur_rev current\n";
    return ("version skew", $cur_rev) ;
  }
  local *F;
  die "Couldn't open $file for wriring: $!" unless open F, ">", $file;
  print F $newtext;
  close F;
  my $remark = remark($req);
  system("ci -u -m'$remark' $file 2>/dev/null < /dev/null");
  Log($req, $page, $req->data->{significant});
  return 1;
}

sub remark {
  my $req = shift;
  my $user = $req->user;
  $user ? "by $user" : $req->{D}{REMOTE_ADDR};
}

sub mkdir_for {
  my $page = shift;
  my @components = split m|/|, $page;
  pop @components;  # pop off filename
  my $DIR = $PAGE;
  for (@components) {
    $DIR .= "/$_";
    warn "$page: mkdir $DIR\n";
    unless (mkdir $DIR, 0777) {
      die "Couldn't create directory $DIR for page $page: $!"
        unless $! == EEXIST;
    }
  }
}

sub fill_template {
  my ($tmpl_name, $req)= @_;
  my $t = template($tmpl_name);
  my $d = $req->data;
  my $v = {REQ => \$req,
           PAGE => $req->page,
           TEXT => scalar(contents($req->page)),
           HTML => scalar(html($req)),
           VAR => $req->data,
          };
  $t->fill_in(HASH => $v);
}

my %last_template;
sub template {
  my $tmpl = shift;
  my $tmpl_file = "$ROOT/templates/$tmpl.tmpl";
  my $lastmod = (stat $tmpl_file)[9];
  if ($lastmod > $last_template{$tmpl}{LAST}) {
    warn "Template '$tmpl' is old; refreshing.\n";
    my $t = Text::Template->new(TYPE => 'FILE',
                                SOURCE => $tmpl_file,)
      or croak "Couldn't build template from $tmpl_file: $Text::Template::ERROR";
    $last_template{$tmpl} = {TMPL => $t, LAST => time() };
  }
  $last_template{$tmpl}{TMPL};
}

sub bottom_menu {
  my $req = shift;
  my @links;
  push @links, reqlink($req, "Edit This Page", "Edit"),
               reqlink($req, "Last Change", "LastChange"),
               reqlink($req, "Change History", "History");
  if ($req->user) {
    push @links, $req->user;
  } else {
    my $return = url_escape(requrl($req));
    my $qs = makeqs(return => $return);
    push @links, qq{<a href="/setcookie.cgi$qs">Log In</a>};
  }
                ;
  join " | ", @links;
}

sub top_menu {
  my $req = shift;
  my @links;
  push @links, reqlink($req, "HomePage", "View", "HomePage");
  push @links, reqlink($req, "RecentChanges", "View", "RecentChanges");
  join " | ", @links;
}

sub reqlink {
  my ($req, $label, $action, $page, $data) = @_;
  $page ||= $req->page;
  $data ||= {};
  my ($host, $port, $path) = @{$req->data}{'SERVER_NAME', 'SERVER_PORT', 'SCRIPT_NAME'};
  $port = ($port == 80 ? "" : ":$port");
  my %data = %$data;
  delete $data{a};  # Overrides the action
  $data{a} = $action unless $action eq "View"; 
  my $qs = makeqs(%data);
  join "", qq{<a href="}, requrl($req, $action, $page, $data), qq{">}, $label, qq{</a>};
}

sub requrl {
  my ($req, $action, $page, $data) = @_;
  $page ||= $req->page;
  $data ||= {};
  $action ||= $req->action;
  my ($host, $port, $path) = @{$req->data}{'SERVER_NAME', 'SERVER_PORT', 'SCRIPT_NAME'};
  $port = ($port == 80 ? "" : ":$port");
  my %data = %$data;
  delete $data{a};  # Overrides the action
  $data{a} = $action unless $action eq "View"; 
  my $qs = makeqs(%data);
  qq{http://$host$port$path/$page/$qs};
}

sub makeqs {
  my %d = @_;
  return "" unless %d;
  my $qs = "?";
  for my $k (keys %d) {
    $qs .= url_escape($k);
    $qs .= "=";
    $qs .= url_escape($d{$k});
    $qs .= "&";
  }
  chop $qs;
  $qs;
}

sub url_escape {
  my $s = shift;
  $s =~ s/([^\w :\/.-])/"%" . sprintf("%2x", $1)/ge;
  $s =~ tr/ /+/;
  $s;
}

sub html_escape {
  my $text = shift;
  my %e = ('>' => 'gt', '<' => 'lt', '&' => 'amp');
  $text =~ s/([<>&])/&$e{$1};/g;
  $text;
}

sub Log {
  my ($req, $page, $significant) = @_;
  unless (fileno LOG) {
    open LOGs, ">>", "$ROOT/logs/Log.sig" or die "Couldn't open $ROOT/Log.sig: $!";
    open LOGm, ">>", "$ROOT/logs/Log.min" or die "Couldn't open $ROOT/Log.min: $!";
    my $ofh = select STDOUT;
    for ('LOGs', 'LOGm') {
      select $_; $| = 1;
    }
    select $ofh;
  }

  unlink "$ROOT/logs/Log.sig.cache" if $significant;
  unlink "$ROOT/logs/Log.min.cache";

  my @fields = (time, ord($req->type), $req->user || '-', $req->data->{REMOTE_ADDR}, $page);
  my $rec = join "\cA", @fields;
  print LOGs $rec, "\cB" if $significant;
  print LOGm $rec, "\cB";
}

1;

