

package Wiki::Socket;
use IO::Socket;
use Carp;
sub eprint;

BEGIN {
  @Wiki::Socket::ISA = 'IO::Socket';
}

sub new {
  my $self = shift;
  my $pack = ref($self) || $self;
  my $sock = IO::Socket->new(@_);
  return unless $sock;
  bless $sock => $pack;
}


# Read exactly $amt bytes.  Return data.
# Return undef (or die, if $die is set) on short read or error
sub rd {
  my ($self, $amt, $die) = @_;
  my $needed = $amt;
  my $result = "";
  until ($needed == 0) {
    my $rc = $self->recv(my $buf, $needed);
    my $br = length $buf;
    if ($ENV{DEBUG}) { eprint "Read $br bytes <$buf>\n" }
    if (! defined $rc || $br == 0) {
      die "$die: $!" if $die;
      return;
    }
    $needed -= $br;
    $result .= $buf;
  }
  $result;
}

# Read a netstring.  Return the string.
# Return undef (or die, if $die is set) on short read,
# malformed netstring, or error
sub rdns {
  my ($self, $die) = @_;
  my $size = $self->rdnsn($die);
  my $text = $self->rd($size, $die);
  $self->rdcksemi("$die: netstring was missing terminating semicolon") or return;
  $text;
}

# Read leading :-terminated length part of netstring
sub rdnsn {
  my ($self, $die) = @_;
  my $size = "";
  until ($size =~ /:$/ || length($size) > 7) {
    my $digit = $self->rd(1, $die);
    if (index("0123456789:", $digit) == -1) {
      die "$die: nondigit '$digit' in length portion of netstring" if $die;
      return;
    }
    $size .= $digit;
  }

  unless ($size =~ s/:$//) {
    die "$die: netstring size too large" if $die;
    return;
  }

  $size;
}

# Read a 'netstring hash': a leading number N, indicating
# a number of key-value pairs; followed by 2N netstrings,
# followed by an additional semicolon.
# It is not an error for keys to be repeated;
# the last occurrence of a key wins
sub rdnshash {
  my ($self, $die) = @_;
  my $npairs = $self->rdnsn($die);
  my %hash;
  while ($npairs--) {
    my $k = $self->rdns($die);
    my $v = $self->rdns($die);
    $hash{$k} = $v;
  }
  $self->rdcksemi($die);
  \%hash;
}

# Make sure the next character is a semicolon
# return true or false; or, if $die is set,
# true or die.
sub rdcksemi {
  my ($self, $die) = @_;
  my $semi = $self->rd(1, $die);
  return 1 if $semi eq ";";
  die "$die:" if $die;
  return;
}

sub get_request {
  my $s = shift;
  my $reqfactory = shift;
  my $req =
    eval {
      my $type = $s->rd(1, "get_request type");
      my $page = $s->rdns("get_request page");
      my $data = $s->rdnshash("get_request data");
      $reqfactory->new($s, $type, $page, $data);
    };
  if ($@) {
    $s->close;
    return;
  }
  $req;
}

################################################################
#
# Code for writing to sockets

# Write a string.  Return true on success; false (or die) on failure
sub wr {
  my ($self, $data, $die) = @_;
  my $offset = 0;
  my $remaining = length $data;
  while ($remaining) {
    my $bw = $self->send(substr($data, $offset), 0);
    if ($ENV{DEBUG}) { eprint "Wrote $bw bytes <".substr($data, $offset, $bw).">\n" }
    if (! defined $bw) {
      die "$die: $!" if $die;
      return;
    }
    $offset += $bw;
    $remaining -= $bw;
  }
  return 1;
}

# Write a string as a netstring
sub wrns {
  my ($self, $s, $die) = @_;
  $self->wrnsn(length($s));
  $self->wr($s, $die);
  $self->wrsemi($die);
}

# Write just the length part of a netstring
sub wrnsn {
  my ($self, $n, $die) = @_;
  croak "Malformed netstring length '$n'" if $n =~ /\D/;
  $self->wr($n, $die);
  $self->wr(":", $die);
}

# Write a hash as a netstring-hash
sub wrnshash {
  my ($self, $hash, $die) = @_;
  my $nk = keys %$hash;
  $self->wrnsn($nk);
  while (my ($k, $v) = each %$hash) {
    eval {
      $self->wrns($k, $die) || die "couldn't write netstring hash key '$k': $!";
      $self->wrns($v, $die) || die "couldn't write netstring hash val '$v': $!";
    };
    if ($@) {
      # reset the hash iterator
      $nk = keys %$hash;
      die; # propagate exception
    }
  }
  $self->wrsemi($die);
}

# Write a netstring trailing sequence: A semicolon
sub wrsemi {
  my ($self, $die) = @_;
  $self->wr(";", $die);
}

sub send_request {
  my $s = shift;
  my $req = shift;
  my $type = $req->type;
  if (length($type) != 1) {
    croak "Request type '$type' must be one character long";
  }
  $s->wr($type, "send_request");
  $s->wrns($req->page, "send_request");
  $s->wrnshash($req->data, "send_request");
}

sub read_response {
  my $sock = shift;
  my $data = $sock->rdns("read_response");
  $data;
}

sub eprint {
  my $t = shift;
  $t =~ s/([^[:print:][:space:]])/"#".ord($1)/ge;
  print STDERR $t;
}
1;

